"""Data models for Lidl Plus API responses."""

from dataclasses import dataclass, field
from datetime import datetime


@dataclass
class Store:
    id: str
    name: str
    address: str
    postal_code: str
    locality: str


@dataclass
class Discount:
    description: str
    amount: float


@dataclass
class Item:
    name: str
    price: float
    vat_group: str
    quantity: float = 1.0
    unit_price: float | None = None
    discounts: list[Discount] = field(default_factory=list)


@dataclass
class Receipt:
    id: str
    date: datetime
    store: Store
    items: list[Item]
    total: float
    currency: str
    payment_method: str | None = None
    vat_breakdown: dict[str, tuple[float, float]] = field(default_factory=dict)


@dataclass
class Coupon:
    id: str
    title: str
    start_date: datetime
    end_date: datetime
    is_activated: bool
    description: str = ""
    image_url: str | None = None



"""Unofficial Lidl Plus API client."""

from importlib.metadata import version

from ilidl.client import LidlClient
from ilidl.models import Coupon, Discount, Item, Receipt, Store

__all__ = ["LidlClient", "Coupon", "Discount", "Item", "Receipt", "Store"]
__version__ = version("ilidl")



"""HTTP client for the Lidl Plus API."""

import base64
from datetime import UTC, datetime, timedelta
from typing import Any

import httpx

from ilidl.exceptions import AuthError, ILidlError
from ilidl.models import Coupon, Receipt, Store
from ilidl.parser import parse_receipt_html

DEFAULT_APP_VERSION = "16.45.5"
AUTH_API = "https://accounts.lidl.com"
TICKETS_API = "https://tickets.lidlplus.com/api"
COUPONS_API = "https://coupons.lidlplus.com/app/api"
CLIENT_ID = "LidlPlusNativeClient"


class _RetryTransport(httpx.BaseTransport):
    """Retries requests on connection errors."""

    def __init__(self, retries: int = 3) -> None:
        self._transport = httpx.HTTPTransport()
        self._retries = retries

    def handle_request(self, request: httpx.Request) -> httpx.Response:
        last_exc: httpx.ConnectError | None = None
        for _ in range(self._retries):
            try:
                return self._transport.handle_request(request)
            except httpx.ConnectError as exc:
                last_exc = exc
        if last_exc is not None:
            raise last_exc
        return self._transport.handle_request(request)

    def close(self) -> None:
        self._transport.close()


class LidlClient:
    """Client for the Lidl Plus API.

    Handles token renewal, receipt listing/detail, and coupons.
    """

    def __init__(
        self,
        refresh_token: str,
        country: str,
        language: str,
        app_version: str = DEFAULT_APP_VERSION,
    ) -> None:
        self._refresh_token = refresh_token
        self._country = country.upper()
        self._language = language.lower()
        self._app_version = app_version
        self._access_token: str = ""
        self._token_expires: datetime = datetime.min.replace(
            tzinfo=UTC,
        )
        self._http = httpx.Client(
            transport=_RetryTransport(retries=3),
            timeout=httpx.Timeout(30.0, connect=10.0),
        )

    def close(self) -> None:
        """Close the underlying HTTP client."""
        self._http.close()

    def __enter__(self) -> "LidlClient":
        return self

    def __exit__(self, *args: object) -> None:
        self.close()

    def _renew_token(self) -> None:
        secret = base64.b64encode(
            f"{CLIENT_ID}:secret".encode(),
        ).decode()
        resp = self._http.post(
            f"{AUTH_API}/connect/token",
            headers={
                "Authorization": f"Basic {secret}",
                "Content-Type": "application/x-www-form-urlencoded",
            },
            data={
                "refresh_token": self._refresh_token,
                "grant_type": "refresh_token",
            },
        )
        try:
            resp.raise_for_status()
        except httpx.HTTPStatusError as exc:
            msg = f"Token renewal failed: {resp.status_code} {resp.text}"
            raise AuthError(msg) from exc
        data = resp.json()
        self._access_token = data["access_token"]
        self._token_expires = datetime.now(
            tz=UTC,
        ) + timedelta(seconds=data["expires_in"])
        if "refresh_token" in data:
            self._refresh_token = data["refresh_token"]

    def _auth_headers(self) -> dict[str, str]:
        now = datetime.now(tz=UTC)
        if not self._access_token or now >= self._token_expires:
            self._renew_token()
        return {
            "Authorization": f"Bearer {self._access_token}",
            "App-Version": self._app_version,
            "Operating-System": "iOs",
            "App": "com.lidl.eci.lidl.plus",
            "Accept-Language": self._language,
        }

    def _get(
        self,
        url: str,
        extra_headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        headers = self._auth_headers()
        if extra_headers:
            headers.update(extra_headers)
        resp = self._http.get(url, headers=headers)
        resp.raise_for_status()
        return resp.json()

    def receipts(
        self,
        only_favourite: bool = False,
    ) -> list[Receipt]:
        """Fetch all receipts, paginating automatically."""
        url = f"{TICKETS_API}/v2/{self._country}/tickets"
        first_page = self._get(
            f"{url}?pageNumber=1&onlyFavorite={only_favourite}",
        )
        tickets = first_page["tickets"]
        total_count = first_page["totalCount"]
        page_size = first_page["size"]
        for page in range(2, (total_count // page_size) + 2):
            page_data = self._get(
                f"{url}?pageNumber={page}&onlyFavorite={only_favourite}",
            )
            tickets.extend(page_data["tickets"])
        return [self._ticket_summary_to_receipt(t) for t in tickets]

    def receipt(self, ticket_id: str) -> Receipt:
        """Fetch a single receipt by ID with full detail."""
        url = f"{TICKETS_API}/v3/{self._country}/tickets/{ticket_id}"
        data = self._get(url)
        return self._ticket_detail_to_receipt(data)

    def latest_receipt(self) -> Receipt:
        """Fetch the most recent receipt with full detail."""
        url = f"{TICKETS_API}/v2/{self._country}/tickets"
        first_page = self._get(
            f"{url}?pageNumber=1&onlyFavorite=False",
        )
        tickets = first_page["tickets"]
        if not tickets:
            msg = "No receipts found"
            raise ILidlError(msg)
        return self.receipt(tickets[0]["id"])

    def _coupon_headers(self) -> dict[str, str]:
        return {"Country": self._country}

    def coupons(self) -> list[Coupon]:
        """Fetch all available coupons."""
        url = f"{COUPONS_API}/v2/promotionsList"
        data = self._get(url, self._coupon_headers())
        result: list[Coupon] = []
        for section in data.get("sections", []):
            for p in section.get("promotions", []):
                validity = p.get("validity", {})
                discount = p.get("discount", {})
                desc = discount.get("description", "")
                if discount.get("title"):
                    desc = f"{discount['title']} {desc}".strip()
                result.append(
                    Coupon(
                        id=p["id"],
                        title=p.get("title", ""),
                        description=desc,
                        image_url=p.get("image"),
                        start_date=datetime.fromisoformat(
                            validity["start"],
                        ),
                        end_date=datetime.fromisoformat(
                            validity["end"],
                        ),
                        is_activated=p.get("isActivated", False),
                    )
                )
        return result

    def activate_coupon(self, coupon_id: str) -> None:
        """Activate a coupon by ID."""
        url = f"{COUPONS_API}/v1/promotions/{coupon_id}/activation"
        headers = self._auth_headers()
        headers.update(self._coupon_headers())
        resp = self._http.post(url, headers=headers)
        resp.raise_for_status()

    def deactivate_coupon(self, coupon_id: str) -> None:
        """Deactivate a coupon by ID."""
        url = f"{COUPONS_API}/v2/promotions/{coupon_id}/activation"
        headers = self._auth_headers()
        headers.update(self._coupon_headers())
        resp = self._http.delete(url, headers=headers)
        resp.raise_for_status()

    @staticmethod
    def _ticket_summary_to_receipt(ticket: dict[str, Any]) -> Receipt:
        currency = ticket.get("currency", {})
        return Receipt(
            id=ticket["id"],
            date=datetime.fromisoformat(ticket["date"]),
            store=Store(
                id=ticket.get("storeCode", ""),
                name="",
                address="",
                postal_code="",
                locality="",
            ),
            items=[],
            total=ticket["totalAmount"],
            currency=currency.get("code", ""),
        )

    @staticmethod
    def _ticket_detail_to_receipt(data: dict[str, Any]) -> Receipt:
        store_data = data.get("store", {})
        store = Store(
            id=store_data.get("id", ""),
            name=store_data.get("name", ""),
            address=store_data.get("address", ""),
            postal_code=store_data.get("postalCode", ""),
            locality=store_data.get("locality", ""),
        )
        html = data.get("htmlPrintedReceipt", "")
        parsed = parse_receipt_html(html)
        return Receipt(
            id=data["id"],
            date=datetime.fromisoformat(data["date"]),
            store=store,
            items=parsed.items,
            total=parsed.total,
            currency=data.get("currency", {}).get("code", "GBP"),
            payment_method=parsed.payment_method,
            vat_breakdown=parsed.vat_breakdown,
        )



"""Parse UK Lidl HTML receipts into structured data."""

from __future__ import annotations

import re
from html.parser import HTMLParser

from ilidl.exceptions import ReceiptParseError
from ilidl.models import Discount, Item


class ParsedReceipt:
    """Result from HTML receipt parsing."""

    def __init__(
        self,
        items: list[Item],
        total: float,
        payment_method: str | None,
        vat_breakdown: dict[str, tuple[float, float]],
    ) -> None:
        self.items = items
        self.total = total
        self.payment_method = payment_method
        self.vat_breakdown = vat_breakdown


class _ReceiptHTMLParser(HTMLParser):
    """Extract data attributes from Lidl receipt HTML spans."""

    def __init__(self) -> None:
        super().__init__()
        self.articles: list[dict[str, str]] = []
        self.vat_entries: list[dict[str, str]] = []
        self.tender_description: str = ""
        self._in_bold = False
        self._current_bold_text = ""
        self._pending_discount_desc: str = ""

    def handle_starttag(
        self,
        tag: str,
        attrs: list[tuple[str, str | None]],
    ) -> None:
        attr_dict = dict(attrs)
        css_class = attr_dict.get("class", "")

        if tag == "span" and "article" in css_class:
            article: dict[str, str] = {}
            for key, val in attrs:
                if key.startswith("data-") and val is not None:
                    article[key] = val
            if "data-art-description" in article:
                self.articles.append(article)

        if tag == "span" and "css_bold" in css_class:
            self._in_bold = True
            self._current_bold_text = ""

        if tag == "span":
            tender = attr_dict.get("data-tender-description")
            if tender:
                self.tender_description = tender

            if attr_dict.get("data-tax-type") and "data-tax-base-amount" in attr_dict:
                entry = {k: v for k, v in attrs if v is not None}
                self.vat_entries.append(entry)

    def handle_data(self, data: str) -> None:
        if self._in_bold:
            self._current_bold_text += data

    def handle_endtag(self, tag: str) -> None:
        if tag != "span" or not self._in_bold:
            return
        self._in_bold = False
        text = self._current_bold_text.strip()

        if text in ("TOTAL", "TOTAL DISCOUNT"):
            return

        if text.startswith("Price") or text.startswith("Coupon"):
            self._pending_discount_desc = text
            return

        if self._pending_discount_desc:
            match = re.match(r"-?([\d.]+)", text)
            if match and self.articles:
                amount = float(match.group(1))
                last_art = self.articles[-1]
                discounts: list[dict[str, object]] = last_art.setdefault("_discounts", [])
                discounts.append(
                    {
                        "description": self._pending_discount_desc,
                        "amount": amount,
                    }
                )
            self._pending_discount_desc = ""


def _parse_total(html: str) -> float:
    """Extract total from the bold TOTAL line."""
    pattern = (
        r'class="css_bold">TOTAL</span>'
        r".*?"
        r'class="css_bold">([\d.]+)</span>'
    )
    match = re.search(pattern, html, re.DOTALL)
    if match:
        return float(match.group(1))
    msg = "Could not find TOTAL in receipt"
    raise ReceiptParseError(msg, html)


def parse_receipt_html(html: str) -> ParsedReceipt:
    """Parse a UK Lidl receipt HTML string into structured data.

    Args:
        html: Raw HTML from the Lidl Plus API's htmlPrintedReceipt field.

    Returns:
        ParsedReceipt with items, total, payment method, and VAT breakdown.

    Raises:
        ReceiptParseError: If no items are found or the HTML cannot be parsed.
    """
    parser = _ReceiptHTMLParser()
    parser.feed(html)

    if not parser.articles:
        msg = "No items found in receipt HTML. This parser supports UK Lidl receipts only."
        raise ReceiptParseError(msg, html)

    seen_ids: set[str] = set()
    items: list[Item] = []

    for art in parser.articles:
        art_id = art.get("data-art-id", "")
        desc = art.get("data-art-description", "")
        key = f"{art_id}:{desc}"
        if key in seen_ids:
            continue
        seen_ids.add(key)

        has_quantity = "data-art-quantity" in art
        quantity = float(art["data-art-quantity"]) if has_quantity else 1.0
        unit_price_raw = float(art.get("data-unit-price", "0"))

        if has_quantity:
            price = round(quantity * unit_price_raw, 2)
            unit_price: float | None = unit_price_raw
        else:
            price = unit_price_raw
            unit_price = None

        discounts = [
            Discount(description=d["description"], amount=d["amount"])
            for d in art.get("_discounts", [])
        ]

        items.append(
            Item(
                name=desc,
                price=price,
                vat_group=art.get("data-tax-type", ""),
                quantity=quantity,
                unit_price=unit_price,
                discounts=discounts,
            )
        )

    total = _parse_total(html)

    vat_breakdown: dict[str, tuple[float, float]] = {}
    for entry in parser.vat_entries:
        tax_type = entry.get("data-tax-type", "")
        base = float(entry.get("data-tax-base-amount", "0"))
        amount = float(entry.get("data-tax-amount", "0"))
        vat_breakdown[tax_type] = (base, amount)

    return ParsedReceipt(
        items=items,
        total=total,
        payment_method=parser.tender_description or None,
        vat_breakdown=vat_breakdown,
    )



"""Tests for ilidl config management."""

from pathlib import Path

from ilidl.config import Config


class TestConfig:
    def test_load_missing_file(self, tmp_path):
        config = Config(tmp_path / "config.toml")
        assert config.refresh_token == ""
        assert config.language == "en"
        assert config.country == "GB"

    def test_save_and_load(self, tmp_path):
        path = tmp_path / "config.toml"
        config = Config(path)
        config.refresh_token = "TOKEN123"
        config.language = "de"
        config.country = "DE"
        config.save()

        loaded = Config(path)
        assert loaded.refresh_token == "TOKEN123"
        assert loaded.language == "de"
        assert loaded.country == "DE"

    def test_save_creates_parent_dirs(self, tmp_path):
        path = tmp_path / "nested" / "dir" / "config.toml"
        config = Config(path)
        config.refresh_token = "TOKEN"
        config.save()
        assert path.exists()

    def test_default_path(self):
        config = Config()
        expected = Path.home() / ".config" / "ilidl" / "config.toml"
        assert config.path == expected



"""CLI for ilidl."""

import json
from dataclasses import asdict
from datetime import datetime
from typing import Any

import click
import httpx

from ilidl.client import LidlClient
from ilidl.config import Config


def _get_client() -> LidlClient:
    config = Config()
    if not config.refresh_token:
        click.echo("Not logged in. Run: ilidl login", err=True)
        raise SystemExit(1)
    client = LidlClient(
        refresh_token=config.refresh_token,
        country=config.country,
        language=config.language,
    )
    click.get_current_context().call_on_close(client.close)
    return client


def _json_serial(obj: Any) -> str:
    if isinstance(obj, datetime):
        return obj.isoformat()
    msg = f"Type {type(obj)} not serializable"
    raise TypeError(msg)


@click.group()
def cli() -> None:
    """Lidl Plus CLI."""


@cli.command()
@click.option("--debug", is_flag=True, help="Save debug screenshots")
def login(*, debug: bool) -> None:
    """Authenticate with Lidl Plus."""
    from ilidl.auth import login as do_login

    config = Config()
    language = click.prompt("Language", default=config.language)
    country = click.prompt("Country", default=config.country)
    config.language = language
    config.country = country
    config.save()

    do_login(config, debug=debug)
    click.echo(f"Logged in. Token saved to {config.path}")


@cli.command("receipts")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option(
    "--from",
    "from_date",
    type=click.DateTime(["%Y-%m-%d"]),
    default=None,
)
@click.option(
    "--to",
    "to_date",
    type=click.DateTime(["%Y-%m-%d"]),
    default=None,
)
def receipts_cmd(
    as_json: bool,
    from_date: datetime | None,
    to_date: datetime | None,
) -> None:
    """List receipts."""
    client = _get_client()
    all_receipts = client.receipts()

    if from_date:
        all_receipts = [r for r in all_receipts if r.date.replace(tzinfo=None) >= from_date]
    if to_date:
        all_receipts = [r for r in all_receipts if r.date.replace(tzinfo=None) <= to_date]

    if as_json:
        click.echo(
            json.dumps(
                [asdict(r) for r in all_receipts],
                default=_json_serial,
                indent=2,
            )
        )
        return

    if not all_receipts:
        click.echo("No receipts found.")
        return

    click.echo(f"{'Date':<20} {'Store':<20} {'Items':>5} {'Total':>8}")
    click.echo("-" * 55)
    for r in all_receipts:
        date_str = r.date.strftime("%Y-%m-%d %H:%M")
        store_name = r.store.name or r.store.id
        click.echo(f"{date_str:<20} {store_name:<20} {len(r.items):>5} {r.total:>7.2f}")


@cli.command("receipt")
@click.argument("ticket_id")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def receipt_cmd(ticket_id: str, as_json: bool) -> None:
    """Show receipt detail. Use 'latest' for most recent."""
    client = _get_client()

    receipt = client.latest_receipt() if ticket_id == "latest" else client.receipt(ticket_id)

    if as_json:
        click.echo(
            json.dumps(
                asdict(receipt),
                default=_json_serial,
                indent=2,
            )
        )
        return

    click.echo(f"Receipt: {receipt.id}")
    click.echo(f"Date:    {receipt.date.strftime('%Y-%m-%d %H:%M')}")
    click.echo(f"Store:   {receipt.store.name}, {receipt.store.locality}")
    click.echo()
    click.echo(f"{'Item':<35} {'Qty':>5} {'Price':>8}")
    click.echo("-" * 50)
    for item in receipt.items:
        qty = f"{item.quantity:.2f}" if item.quantity != 1.0 else ""
        click.echo(f"{item.name:<35} {qty:>5} {item.price:>7.2f}")
        for d in item.discounts:
            click.echo(f"  {d.description:<33} {'':<5} {-d.amount:>7.2f}")
    click.echo("-" * 50)
    click.echo(f"{'TOTAL':<35} {'':<5} {receipt.total:>7.2f}")
    if receipt.payment_method:
        click.echo(f"Paid by: {receipt.payment_method}")


@cli.group("coupons")
def coupons_group() -> None:
    """Manage coupons."""


@coupons_group.command("list")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def coupons_list(as_json: bool) -> None:
    """List available coupons."""
    client = _get_client()
    all_coupons = client.coupons()

    if as_json:
        click.echo(
            json.dumps(
                [asdict(c) for c in all_coupons],
                default=_json_serial,
                indent=2,
            )
        )
        return

    if not all_coupons:
        click.echo("No coupons available.")
        return

    click.echo(f"{'Title':<30} {'Valid Until':<12} {'Active':>6}")
    click.echo("-" * 50)
    for c in all_coupons:
        end = c.end_date.strftime("%Y-%m-%d")
        active = "Yes" if c.is_activated else "No"
        click.echo(f"{c.title:<30} {end:<12} {active:>6}")


@coupons_group.command("activate")
@click.argument("coupon_id", required=False)
@click.option(
    "--all",
    "activate_all",
    is_flag=True,
    help="Activate all coupons",
)
def coupons_activate(coupon_id: str | None, activate_all: bool) -> None:
    """Activate a coupon."""
    client = _get_client()
    if activate_all:
        for c in client.coupons():
            if not c.is_activated:
                try:
                    client.activate_coupon(c.id)
                    click.echo(f"Activated: {c.title}")
                except httpx.HTTPStatusError as e:
                    click.echo(
                        f"Skipped: {c.title} ({e.response.status_code})",
                        err=True,
                    )
        return
    if not coupon_id:
        click.echo("Provide a coupon ID or use --all", err=True)
        raise SystemExit(1)
    client.activate_coupon(coupon_id)
    click.echo("Coupon activated.")


@coupons_group.command("deactivate")
@click.argument("coupon_id")
def coupons_deactivate(coupon_id: str) -> None:
    """Deactivate a coupon."""
    client = _get_client()
    client.deactivate_coupon(coupon_id)
    click.echo("Coupon deactivated.")



"""Playwright-based OAuth2 PKCE authentication for Lidl Plus."""

import base64
import contextlib
import getpass
import hashlib
import re
import secrets
from typing import Any

import httpx

from ilidl.config import Config
from ilidl.exceptions import AuthError

AUTH_API = "https://accounts.lidl.com"
CLIENT_ID = "LidlPlusNativeClient"
REDIRECT_URI = "com.lidlplus.app://callback"
SCOPES = "openid profile offline_access lpprofile lpapis"


def _generate_pkce() -> tuple[str, str]:
    """Generate PKCE code_verifier and code_challenge."""
    verifier = secrets.token_urlsafe(64)
    digest = hashlib.sha256(verifier.encode()).digest()
    challenge = base64.urlsafe_b64encode(digest).rstrip(b"=").decode()
    return verifier, challenge


def _build_auth_url(challenge: str, country: str, language: str) -> str:
    """Build the OAuth authorization URL."""
    params = (
        f"client_id={CLIENT_ID}"
        f"&response_type=code"
        f"&scope={SCOPES.replace(' ', '+')}"
        f"&redirect_uri={REDIRECT_URI.replace(':', '%3A').replace('/', '%2F')}"
        f"&code_challenge={challenge}"
        f"&code_challenge_method=S256"
        f"&Country={country.upper()}"
        f"&language={language.lower()}-{country.upper()}"
    )
    return f"{AUTH_API}/connect/authorize?{params}"


def _exchange_code(code: str, verifier: str) -> dict[str, Any]:
    """Exchange authorization code for tokens."""
    secret = base64.b64encode(f"{CLIENT_ID}:secret".encode()).decode()
    resp = httpx.post(
        f"{AUTH_API}/connect/token",
        headers={
            "Authorization": f"Basic {secret}",
            "Content-Type": "application/x-www-form-urlencoded",
        },
        data={
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": REDIRECT_URI,
            "code_verifier": verifier,
        },
        timeout=30,
    )
    try:
        resp.raise_for_status()
    except httpx.HTTPStatusError as exc:
        msg = f"Token exchange failed: {resp.status_code} {resp.text}"
        raise AuthError(msg) from exc
    return resp.json()


def _dbg(debug: bool, page: Any, name: str, msg: str = "") -> None:
    """Save a debug screenshot and optionally print a message."""
    if not debug:
        return
    page.screenshot(path=f"/tmp/ilidl_{name}.png")
    if msg:
        print(f"DEBUG {msg}")


def login(config: Config, *, debug: bool = False) -> str:
    """Run interactive login via phone number. Returns refresh token."""
    try:
        from playwright.sync_api import sync_playwright
    except ImportError as e:
        msg = "Playwright is required for login. Install with: uv pip install 'ilidl[auth]'"
        raise AuthError(msg) from e

    verifier, challenge = _generate_pkce()
    url = _build_auth_url(challenge, config.country, config.language)

    phone = input("Phone number (local, e.g. 7415148717): ").strip()
    password = getpass.getpass("Password: ")

    with sync_playwright() as p:
        browser = p.chromium.launch(headless=True)
        page = browser.new_page()

        code: str = ""

        def handle_request(request):
            nonlocal code
            req_url = request.url
            if debug:
                print(f"DEBUG request: {req_url[:120]}")
            if req_url.startswith("com.lidlplus.app://"):
                match = re.search(r"code=([^&]+)", req_url)
                if match:
                    code = match.group(1)
                    if debug:
                        print(f"DEBUG got code: {code}")

        def handle_response(response):
            nonlocal code
            # Check for 302 redirects to the app callback
            if code:
                return
            location = response.headers.get("location", "")
            if location.startswith("com.lidlplus.app://"):
                match = re.search(r"code=([^&]+)", location)
                if match:
                    code = match.group(1)
                    if debug:
                        print(f"DEBUG got code from redirect: {code}")

        page.on("request", handle_request)
        page.on("response", handle_response)

        page.goto(url)
        page.wait_for_load_state("networkidle")
        _dbg(debug, page, "01_welcome", f"url: {page.url}")

        # Click "Log in" on the welcome page
        page.locator("button:has-text('Log in'), a:has-text('Log in')").first.click(timeout=5000)
        page.wait_for_load_state("networkidle")
        _dbg(debug, page, "02_login_form", "on login form")

        # Switch to phone number login
        page.locator("button:has-text('phone number'), a:has-text('phone number')").first.click(
            timeout=5000
        )
        page.wait_for_load_state("networkidle")
        _dbg(debug, page, "03_phone_form", f"url: {page.url}")

        # Fill local phone number (country code is pre-selected)
        phone_field = page.locator(
            "input[type='tel'], input[name='PhoneNumber'], input[name='phoneNumber']"
        ).first
        phone_field.fill(phone)

        # Fill password
        page.locator("input[type='password']").first.fill(password)
        _dbg(debug, page, "04_filled")

        # Submit and wait for navigation or page change
        login_url = page.url
        page.locator("button:has-text('Log in'), button[type='submit']").first.click()

        # Wait for URL to change or for an error/verification
        # element to appear
        with contextlib.suppress(TimeoutError):
            page.wait_for_function(
                f"window.location.href !== '{login_url}' || "
                "document.querySelector("
                '  \'[name="VerificationCode"],'
                "  .alert-danger,"
                "  .error-message'"
                ")",
                timeout=15000,
            )
        page.wait_for_load_state("networkidle")
        _dbg(debug, page, "05_after_submit", f"url: {page.url}")

        # Check for error message
        if not code:
            error_el = page.locator(
                ".alert-danger, .error-message, [class*='error'], [class*='Error']"
            )
            if error_el.first.is_visible(timeout=2000):
                error_text = error_el.first.text_content()
                _dbg(debug, page, "05b_error")
                browser.close()
                msg = f"Login failed: {error_text}"
                raise AuthError(msg)

        # Check for verification code prompt
        if not code:
            verify_field = page.locator(
                "[name='VerificationCode'], input[name='code'], input[inputmode='numeric']"
            )
            try:
                if verify_field.first.is_visible(timeout=10000):
                    _dbg(debug, page, "06_verify_prompt")
                    otp = input("Enter verification code: ").strip()
                    verify_field.first.fill(otp)
                    page.locator(
                        "button[type='submit'], "
                        "button:has-text('Submit'), "
                        "button:has-text('Verify')"
                    ).first.click()

                    # Wait for redirect after verify
                    with contextlib.suppress(TimeoutError):
                        page.wait_for_url("**/callback*", timeout=15000)
                    page.wait_for_timeout(3000)
                    _dbg(debug, page, "07_after_verify", f"url: {page.url}")
            except Exception as exc:
                if debug:
                    print(f"DEBUG verify error: {exc}")
                _dbg(debug, page, "08_error")

        browser.close()

    if not code:
        msg = "Login failed: could not extract authorization code"
        raise AuthError(msg)

    tokens = _exchange_code(code, verifier)
    config.refresh_token = tokens["refresh_token"]
    config.save()
    return tokens["refresh_token"]



"""Tests for ilidl auth module."""

from unittest.mock import MagicMock, patch

import httpx
import pytest

from ilidl.auth import _build_auth_url, _exchange_code, _generate_pkce
from ilidl.exceptions import AuthError


class TestGeneratePkce:
    def test_returns_verifier_and_challenge(self):
        verifier, challenge = _generate_pkce()
        assert isinstance(verifier, str)
        assert isinstance(challenge, str)
        assert len(verifier) > 0
        assert len(challenge) > 0

    def test_challenge_differs_from_verifier(self):
        verifier, challenge = _generate_pkce()
        assert verifier != challenge

    def test_generates_unique_values(self):
        v1, _ = _generate_pkce()
        v2, _ = _generate_pkce()
        assert v1 != v2

    def test_challenge_is_base64url_without_padding(self):
        _, challenge = _generate_pkce()
        assert "+" not in challenge
        assert "/" not in challenge
        assert "=" not in challenge


class TestBuildAuthUrl:
    def test_contains_required_params(self):
        url = _build_auth_url("test_challenge", "gb", "en")
        assert "client_id=LidlPlusNativeClient" in url
        assert "response_type=code" in url
        assert "code_challenge=test_challenge" in url
        assert "code_challenge_method=S256" in url

    def test_uppercases_country(self):
        url = _build_auth_url("c", "gb", "en")
        assert "Country=GB" in url
        assert "language=en-GB" in url

    def test_encodes_redirect_uri(self):
        url = _build_auth_url("c", "gb", "en")
        assert "com.lidlplus.app%3A%2F%2Fcallback" in url

    def test_base_url(self):
        url = _build_auth_url("c", "gb", "en")
        assert url.startswith("https://accounts.lidl.com/connect/authorize?")


class TestExchangeCode:
    def test_returns_token_data(self):
        mock_resp = MagicMock(spec=httpx.Response)
        mock_resp.status_code = 200
        mock_resp.json.return_value = {
            "access_token": "at",
            "refresh_token": "rt",
        }
        mock_resp.raise_for_status.return_value = None
        with patch("ilidl.auth.httpx.post", return_value=mock_resp):
            result = _exchange_code("code", "verifier")
        assert result["access_token"] == "at"
        assert result["refresh_token"] == "rt"

    def test_raises_auth_error_on_failure(self):
        mock_resp = MagicMock(spec=httpx.Response)
        mock_resp.status_code = 400
        mock_resp.text = "invalid_grant"
        mock_resp.raise_for_status.side_effect = httpx.HTTPStatusError(
            "400",
            request=MagicMock(),
            response=mock_resp,
        )
        with (
            patch("ilidl.auth.httpx.post", return_value=mock_resp),
            pytest.raises(AuthError, match="Token exchange failed: 400"),
        ):
            _exchange_code("code", "verifier")

    def test_sends_correct_payload(self):
        mock_resp = MagicMock(spec=httpx.Response)
        mock_resp.status_code = 200
        mock_resp.json.return_value = {"access_token": "at"}
        mock_resp.raise_for_status.return_value = None
        with patch("ilidl.auth.httpx.post", return_value=mock_resp) as mock_post:
            _exchange_code("my_code", "my_verifier")
        call_kwargs = mock_post.call_args.kwargs
        assert call_kwargs["data"]["code"] == "my_code"
        assert call_kwargs["data"]["code_verifier"] == "my_verifier"
        assert call_kwargs["data"]["grant_type"] == "authorization_code"



"""Config file management for ilidl."""

import tomllib
from pathlib import Path

import tomli_w

_DEFAULT_PATH = Path.home() / ".config" / "ilidl" / "config.toml"


class Config:
    def __init__(self, path: Path | None = None) -> None:
        self.path = path or _DEFAULT_PATH
        self.refresh_token: str = ""
        self.language: str = "en"
        self.country: str = "GB"
        self._load()

    def _load(self) -> None:
        if not self.path.exists():
            return
        with open(self.path, "rb") as f:
            data = tomllib.load(f)
        auth = data.get("auth", {})
        self.refresh_token = auth.get("refresh_token", "")
        account = data.get("account", {})
        self.language = account.get("language", "en")
        self.country = account.get("country", "GB")

    def save(self) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        data = {
            "auth": {"refresh_token": self.refresh_token},
            "account": {
                "language": self.language,
                "country": self.country,
            },
        }
        with open(self.path, "wb") as f:
            tomli_w.dump(data, f)



"""Tests for ilidl API client."""

import json
from pathlib import Path
from unittest.mock import MagicMock

import httpx
import pytest

from ilidl.client import LidlClient
from ilidl.exceptions import AuthError, ILidlError
from ilidl.models import Receipt

FIXTURE = Path(__file__).parent / "fixtures" / "receipt_gb_sample.json"


def _mock_response(status_code: int = 200, json_data: dict | None = None, text: str = ""):
    resp = MagicMock(spec=httpx.Response)
    resp.status_code = status_code
    resp.text = text
    resp.json.return_value = json_data or {}
    if status_code >= 400:
        resp.raise_for_status.side_effect = httpx.HTTPStatusError(
            f"{status_code}",
            request=MagicMock(),
            response=resp,
        )
    else:
        resp.raise_for_status.return_value = None
    return resp


def _token_json(include_refresh: bool = True) -> dict:
    data = {"access_token": "test_access_token", "expires_in": 1200}
    if include_refresh:
        data["refresh_token"] = "new_refresh_token"
    return data


def _tickets_json(
    tickets: list | None = None,
    total_count: int | None = None,
    page_size: int = 25,
) -> dict:
    if tickets is None:
        tickets = [
            {
                "id": "receipt123",
                "date": "2026-03-18T19:26:50+00:00",
                "totalAmount": 12.38,
                "currency": {"code": "GBP", "symbol": "\u00a3"},
                "storeCode": "GB1459",
                "articlesCount": 7,
            }
        ]
    if total_count is None:
        total_count = len(tickets)
    return {"page": 1, "size": page_size, "totalCount": total_count, "tickets": tickets}


def _detail_json() -> dict:
    return json.loads(FIXTURE.read_text())


def _make_client() -> LidlClient:
    client = LidlClient(
        refresh_token="test_refresh",
        country="GB",
        language="en",
    )
    client._http = MagicMock(spec=httpx.Client)
    return client


def _setup_token(client: LidlClient, include_refresh: bool = True) -> None:
    """Configure mock to return a valid token on the next POST call."""
    client._http.post.return_value = _mock_response(
        json_data=_token_json(include_refresh),
    )


class TestTokenRenewal:
    def test_renews_token_on_first_call(self):
        client = _make_client()
        _setup_token(client)
        headers = client._auth_headers()
        assert headers["Authorization"] == "Bearer test_access_token"
        client._http.post.assert_called_once()

    def test_reuses_token_when_not_expired(self):
        client = _make_client()
        _setup_token(client)
        client._auth_headers()
        client._auth_headers()
        assert client._http.post.call_count == 1

    def test_renewal_failure_raises_auth_error(self):
        client = _make_client()
        client._http.post.return_value = _mock_response(
            status_code=401,
            text="invalid_grant",
        )
        with pytest.raises(AuthError, match="Token renewal failed: 401"):
            client._auth_headers()

    def test_renewal_updates_refresh_token(self):
        client = _make_client()
        _setup_token(client, include_refresh=True)
        client._auth_headers()
        assert client._refresh_token == "new_refresh_token"

    def test_renewal_keeps_old_refresh_token_when_not_returned(self):
        client = _make_client()
        _setup_token(client, include_refresh=False)
        client._auth_headers()
        assert client._refresh_token == "test_refresh"

    def test_auth_headers_contain_required_fields(self):
        client = _make_client()
        _setup_token(client)
        headers = client._auth_headers()
        assert "Authorization" in headers
        assert headers["App-Version"] == "16.45.5"
        assert headers["Operating-System"] == "iOs"
        assert headers["Accept-Language"] == "en"

    def test_custom_app_version(self):
        client = LidlClient(
            refresh_token="test",
            country="GB",
            language="en",
            app_version="17.0.0",
        )
        client._http = MagicMock(spec=httpx.Client)
        _setup_token(client)
        headers = client._auth_headers()
        assert headers["App-Version"] == "17.0.0"


class TestGet:
    def test_get_passes_extra_headers(self):
        client = _make_client()
        _setup_token(client)
        client._http.get.return_value = _mock_response(json_data={"ok": True})
        client._get("https://example.com", extra_headers={"Country": "GB"})
        call_headers = client._http.get.call_args.kwargs["headers"]
        assert call_headers["Country"] == "GB"
        assert "Authorization" in call_headers

    def test_get_raises_on_http_error(self):
        client = _make_client()
        _setup_token(client)
        client._http.get.return_value = _mock_response(status_code=500)
        with pytest.raises(httpx.HTTPStatusError):
            client._get("https://example.com")


class TestReceipts:
    def test_latest_receipt(self):
        client = _make_client()
        _setup_token(client)
        client._http.get.side_effect = [
            _mock_response(json_data=_tickets_json()),
            _mock_response(json_data=_detail_json()),
        ]
        receipt = client.latest_receipt()
        assert isinstance(receipt, Receipt)
        assert receipt.total == 12.38
        assert len(receipt.items) == 7

    def test_latest_receipt_no_receipts_raises(self):
        client = _make_client()
        _setup_token(client)
        client._http.get.return_value = _mock_response(
            json_data=_tickets_json(tickets=[], total_count=0),
        )
        with pytest.raises(ILidlError, match="No receipts found"):
            client.latest_receipt()

    def test_receipt_by_id(self):
        client = _make_client()
        _setup_token(client)
        client._http.get.return_value = _mock_response(json_data=_detail_json())
        receipt = client.receipt("receipt123")
        assert receipt.id == "13001459742026031831079"
        assert receipt.store.name == "Llandaff"

    def test_receipts_single_page(self):
        client = _make_client()
        _setup_token(client)
        client._http.get.return_value = _mock_response(json_data=_tickets_json())
        result = client.receipts()
        assert len(result) == 1
        assert result[0].id == "receipt123"

    def test_receipts_pagination(self):
        """Two pages: page 1 has 2 tickets, page 2 has 1 ticket."""
        ticket_a = {
            "id": "a",
            "date": "2026-03-18T10:00:00+00:00",
            "totalAmount": 5.00,
            "currency": {"code": "GBP"},
            "storeCode": "GB1",
        }
        ticket_b = {**ticket_a, "id": "b"}
        ticket_c = {**ticket_a, "id": "c"}
        page1 = _mock_response(
            json_data={
                "page": 1,
                "size": 2,
                "totalCount": 3,
                "tickets": [ticket_a, ticket_b],
            }
        )
        page2 = _mock_response(
            json_data={
                "page": 2,
                "size": 2,
                "totalCount": 3,
                "tickets": [ticket_c],
            }
        )
        client = _make_client()
        _setup_token(client)
        client._http.get.side_effect = [page1, page2]
        result = client.receipts()
        assert len(result) == 3
        assert [r.id for r in result] == ["a", "b", "c"]

    def test_receipts_only_favourite(self):
        client = _make_client()
        _setup_token(client)
        client._http.get.return_value = _mock_response(json_data=_tickets_json())
        client.receipts(only_favourite=True)
        url = client._http.get.call_args_list[0].args[0]
        assert "onlyFavorite=True" in url


class TestCoupons:
    def test_coupons_parses_sections(self):
        client = _make_client()
        _setup_token(client)
        coupon_data = {
            "sections": [
                {
                    "promotions": [
                        {
                            "id": "c1",
                            "title": "Free Pizza",
                            "validity": {
                                "start": "2026-03-01T00:00:00+00:00",
                                "end": "2026-03-31T23:59:59+00:00",
                            },
                            "discount": {
                                "title": "50% off",
                                "description": "on all pizzas",
                            },
                            "image": "https://example.com/pizza.jpg",
                            "isActivated": True,
                        }
                    ]
                }
            ]
        }
        client._http.get.return_value = _mock_response(json_data=coupon_data)
        coupons = client.coupons()
        assert len(coupons) == 1
        assert coupons[0].id == "c1"
        assert coupons[0].title == "Free Pizza"
        assert coupons[0].description == "50% off on all pizzas"
        assert coupons[0].image_url == "https://example.com/pizza.jpg"
        assert coupons[0].is_activated is True

    def test_coupons_empty_sections(self):
        client = _make_client()
        _setup_token(client)
        client._http.get.return_value = _mock_response(
            json_data={"sections": []},
        )
        assert client.coupons() == []

    def test_activate_coupon(self):
        client = _make_client()
        _setup_token(client)
        client._http.post.side_effect = [
            _mock_response(json_data=_token_json()),
            _mock_response(),
        ]
        client.activate_coupon("c1")
        assert client._http.post.call_count == 2
        activate_call = client._http.post.call_args_list[1]
        assert "c1/activation" in activate_call.args[0]
        assert activate_call.kwargs["headers"]["Country"] == "GB"

    def test_activate_coupon_http_error(self):
        client = _make_client()
        _setup_token(client)
        client._http.post.side_effect = [
            _mock_response(json_data=_token_json()),
            _mock_response(status_code=409),
        ]
        with pytest.raises(httpx.HTTPStatusError):
            client.activate_coupon("c1")

    def test_deactivate_coupon(self):
        client = _make_client()
        _setup_token(client)
        client._http.delete.return_value = _mock_response()
        client.deactivate_coupon("c1")
        url = client._http.delete.call_args.args[0]
        assert "c1/activation" in url

    def test_deactivate_coupon_http_error(self):
        client = _make_client()
        _setup_token(client)
        client._http.delete.return_value = _mock_response(status_code=404)
        with pytest.raises(httpx.HTTPStatusError):
            client.deactivate_coupon("c1")


class TestTicketConversion:
    def test_summary_to_receipt(self):
        ticket = {
            "id": "r1",
            "date": "2026-03-18T19:26:50+00:00",
            "totalAmount": 12.38,
            "currency": {"code": "GBP", "symbol": "\u00a3"},
            "storeCode": "GB1459",
        }
        receipt = LidlClient._ticket_summary_to_receipt(ticket)
        assert receipt.id == "r1"
        assert receipt.total == 12.38
        assert receipt.currency == "GBP"
        assert receipt.store.id == "GB1459"
        assert receipt.items == []

    def test_summary_missing_currency(self):
        ticket = {
            "id": "r1",
            "date": "2026-03-18T19:26:50+00:00",
            "totalAmount": 5.00,
        }
        receipt = LidlClient._ticket_summary_to_receipt(ticket)
        assert receipt.currency == ""

    def test_detail_to_receipt(self):
        data = json.loads(FIXTURE.read_text())
        receipt = LidlClient._ticket_detail_to_receipt(data)
        assert receipt.store.name == "Llandaff"
        assert receipt.store.locality == "Cardiff"
        assert len(receipt.items) == 7
        assert receipt.payment_method == "CARD"
        assert "A" in receipt.vat_breakdown
        assert "B" in receipt.vat_breakdown


class TestClientLifecycle:
    def test_context_manager(self):
        client = _make_client()
        with client:
            pass
        client._http.close.assert_called_once()

    def test_close(self):
        client = _make_client()
        client.close()
        client._http.close.assert_called_once()



"""Tests for ilidl HTML receipt parser."""

import json
from pathlib import Path

import pytest

from ilidl.exceptions import ReceiptParseError
from ilidl.parser import parse_receipt_html

FIXTURE = Path(__file__).parent / "fixtures" / "receipt_gb_sample.json"


@pytest.fixture
def sample_html():
    data = json.loads(FIXTURE.read_text())
    return data["htmlPrintedReceipt"]


class TestParseItems:
    def test_item_count(self, sample_html):
        result = parse_receipt_html(sample_html)
        assert len(result.items) == 7

    def test_simple_item(self, sample_html):
        result = parse_receipt_html(sample_html)
        beans = next(i for i in result.items if "Beans" in i.name)
        assert beans.name == "Heinz Beans Standard"
        assert beans.price == 2.74
        assert beans.vat_group == "A"
        assert beans.quantity == 1.0

    def test_weight_item(self, sample_html):
        result = parse_receipt_html(sample_html)
        carrots = next(i for i in result.items if "Carrots" in i.name)
        assert carrots.name == "Loose Carrots 0082755"
        assert carrots.price == 0.17
        assert carrots.quantity == 0.24
        assert carrots.unit_price == 0.69

    def test_item_with_discount(self, sample_html):
        result = parse_receipt_html(sample_html)
        handwash = next(i for i in result.items if "Handwash" in i.name)
        assert len(handwash.discounts) == 1
        assert handwash.discounts[0].description == "Price Cut"
        assert handwash.discounts[0].amount == 0.04


class TestParseTotals:
    def test_total(self, sample_html):
        result = parse_receipt_html(sample_html)
        assert result.total == 12.38

    def test_payment_method(self, sample_html):
        result = parse_receipt_html(sample_html)
        assert result.payment_method == "CARD"


class TestParseVat:
    def test_vat_breakdown(self, sample_html):
        result = parse_receipt_html(sample_html)
        assert result.vat_breakdown["A"] == (5.16, 0.00)
        assert result.vat_breakdown["B"] == (7.22, 1.20)


class TestParseError:
    def test_empty_html_raises(self):
        with pytest.raises(ReceiptParseError) as exc_info:
            parse_receipt_html("<html><body></body></html>")
        assert exc_info.value.raw_html == "<html><body></body></html>"

    def test_no_items_raises(self):
        with pytest.raises(ReceiptParseError):
            parse_receipt_html("<html><body><pre></pre></body></html>")

    def test_missing_total_raises(self):
        html = (
            "<html><body><pre>"
            '<span class="article" data-art-id="1" '
            'data-unit-price="2.00" data-tax-type="A" '
            'data-art-description="Widget">Widget 2.00</span>'
            "</pre></body></html>"
        )
        with pytest.raises(ReceiptParseError, match="Could not find TOTAL"):
            parse_receipt_html(html)



"""Tests for ilidl CLI."""

import json
from datetime import datetime
from unittest.mock import MagicMock, patch

import httpx
import pytest
from click.testing import CliRunner

from ilidl.cli import _json_serial, cli
from ilidl.models import Coupon, Discount, Item, Receipt, Store


def _sample_receipt(**overrides):
    defaults = {
        "id": "receipt123",
        "date": datetime(2026, 3, 18, 19, 26, 50),
        "store": Store(
            id="GB1459",
            name="Llandaff",
            address="Station Road",
            postal_code="CF14 2FB",
            locality="Cardiff",
        ),
        "items": [
            Item(name="Heinz Beans", price=2.74, vat_group="A"),
            Item(
                name="Handwash",
                price=0.59,
                vat_group="B",
                discounts=[Discount(description="Price Cut", amount=0.04)],
            ),
        ],
        "total": 3.29,
        "currency": "GBP",
        "payment_method": "CARD",
    }
    defaults.update(overrides)
    return Receipt(**defaults)


def _sample_coupon(**overrides):
    defaults = {
        "id": "coupon123",
        "title": "Free Pizza",
        "start_date": datetime(2026, 3, 1),
        "end_date": datetime(2026, 3, 31),
        "is_activated": False,
        "description": "Buy 1 get 1 free",
    }
    defaults.update(overrides)
    return Coupon(**defaults)


class TestJsonSerial:
    def test_serialises_datetime(self):
        dt = datetime(2026, 3, 18, 19, 26, 50)
        assert _json_serial(dt) == "2026-03-18T19:26:50"

    def test_raises_for_unsupported_type(self):
        with pytest.raises(TypeError, match="not serializable"):
            _json_serial(set())


class TestGetClient:
    @patch("ilidl.cli.Config")
    def test_not_logged_in(self, mock_config_cls):
        mock_config_cls.return_value.refresh_token = ""
        runner = CliRunner()
        result = runner.invoke(cli, ["receipts"])
        assert result.exit_code == 1
        assert "Not logged in" in result.output


class TestReceiptCommand:
    @patch("ilidl.cli._get_client")
    def test_receipt_latest_json(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.latest_receipt.return_value = _sample_receipt()
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipt", "latest", "--json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert data["id"] == "receipt123"
        assert data["total"] == 3.29

    @patch("ilidl.cli._get_client")
    def test_receipt_latest_table(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.latest_receipt.return_value = _sample_receipt()
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipt", "latest"])
        assert result.exit_code == 0
        assert "Llandaff" in result.output
        assert "Heinz Beans" in result.output
        assert "Price Cut" in result.output
        assert "Paid by: CARD" in result.output
        assert "TOTAL" in result.output

    @patch("ilidl.cli._get_client")
    def test_receipt_by_id(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.receipt.return_value = _sample_receipt()
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipt", "abc123"])
        assert result.exit_code == 0
        mock_client.receipt.assert_called_once_with("abc123")

    @patch("ilidl.cli._get_client")
    def test_receipt_table_shows_quantity(self, mock_get_client):
        receipt = _sample_receipt(
            items=[
                Item(
                    name="Loose Carrots",
                    price=0.17,
                    vat_group="A",
                    quantity=0.24,
                    unit_price=0.69,
                ),
            ],
            total=0.17,
        )
        mock_client = MagicMock()
        mock_client.latest_receipt.return_value = receipt
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipt", "latest"])
        assert result.exit_code == 0
        assert "0.24" in result.output

    @patch("ilidl.cli._get_client")
    def test_receipt_table_no_payment_method(self, mock_get_client):
        receipt = _sample_receipt(payment_method=None)
        mock_client = MagicMock()
        mock_client.latest_receipt.return_value = receipt
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipt", "latest"])
        assert result.exit_code == 0
        assert "Paid by" not in result.output


class TestReceiptsCommand:
    @patch("ilidl.cli._get_client")
    def test_receipts_table(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.receipts.return_value = [_sample_receipt()]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipts"])
        assert result.exit_code == 0
        assert "2026-03-18" in result.output
        assert "Llandaff" in result.output
        assert "3.29" in result.output

    @patch("ilidl.cli._get_client")
    def test_receipts_json(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.receipts.return_value = [_sample_receipt()]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipts", "--json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert len(data) == 1
        assert data[0]["id"] == "receipt123"

    @patch("ilidl.cli._get_client")
    def test_receipts_empty(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.receipts.return_value = []
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipts"])
        assert result.exit_code == 0
        assert "No receipts found" in result.output

    @patch("ilidl.cli._get_client")
    def test_receipts_date_filter_from(self, mock_get_client):
        old = _sample_receipt(
            id="old",
            date=datetime(2026, 1, 1, 10, 0, 0),
        )
        new = _sample_receipt(
            id="new",
            date=datetime(2026, 3, 18, 19, 26, 50),
        )
        mock_client = MagicMock()
        mock_client.receipts.return_value = [old, new]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipts", "--from", "2026-03-01", "--json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert len(data) == 1
        assert data[0]["id"] == "new"

    @patch("ilidl.cli._get_client")
    def test_receipts_date_filter_to(self, mock_get_client):
        old = _sample_receipt(
            id="old",
            date=datetime(2026, 1, 1, 10, 0, 0),
        )
        new = _sample_receipt(
            id="new",
            date=datetime(2026, 3, 18, 19, 26, 50),
        )
        mock_client = MagicMock()
        mock_client.receipts.return_value = [old, new]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipts", "--to", "2026-02-01", "--json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert len(data) == 1
        assert data[0]["id"] == "old"

    @patch("ilidl.cli._get_client")
    def test_receipts_store_falls_back_to_id(self, mock_get_client):
        receipt = _sample_receipt(
            store=Store(id="GB1459", name="", address="", postal_code="", locality=""),
        )
        mock_client = MagicMock()
        mock_client.receipts.return_value = [receipt]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["receipts"])
        assert result.exit_code == 0
        assert "GB1459" in result.output


class TestCouponsCommand:
    @patch("ilidl.cli._get_client")
    def test_coupons_list_table(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.coupons.return_value = [_sample_coupon()]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["coupons", "list"])
        assert result.exit_code == 0
        assert "Free Pizza" in result.output
        assert "2026-03-31" in result.output
        assert "No" in result.output

    @patch("ilidl.cli._get_client")
    def test_coupons_list_json(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.coupons.return_value = [_sample_coupon()]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["coupons", "list", "--json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert len(data) == 1
        assert data[0]["title"] == "Free Pizza"

    @patch("ilidl.cli._get_client")
    def test_coupons_list_empty(self, mock_get_client):
        mock_client = MagicMock()
        mock_client.coupons.return_value = []
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["coupons", "list"])
        assert result.exit_code == 0
        assert "No coupons available" in result.output

    @patch("ilidl.cli._get_client")
    def test_coupons_activate_by_id(self, mock_get_client):
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["coupons", "activate", "c1"])
        assert result.exit_code == 0
        assert "Coupon activated" in result.output
        mock_client.activate_coupon.assert_called_once_with("c1")

    @patch("ilidl.cli._get_client")
    def test_coupons_activate_no_id_no_all(self, mock_get_client):
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["coupons", "activate"])
        assert result.exit_code == 1
        assert "Provide a coupon ID" in result.output

    @patch("ilidl.cli._get_client")
    def test_coupons_activate_all(self, mock_get_client):
        inactive = _sample_coupon(id="c1", is_activated=False)
        active = _sample_coupon(id="c2", is_activated=True)
        mock_client = MagicMock()
        mock_client.coupons.return_value = [inactive, active]
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["coupons", "activate", "--all"])
        assert result.exit_code == 0
        assert "Activated" in result.output
        mock_client.activate_coupon.assert_called_once_with("c1")

    @patch("ilidl.cli._get_client")
    def test_coupons_activate_all_skips_on_error(self, mock_get_client):
        coupon = _sample_coupon(id="c1", is_activated=False)
        mock_client = MagicMock()
        mock_client.coupons.return_value = [coupon]
        error_resp = MagicMock(status_code=409)
        mock_client.activate_coupon.side_effect = httpx.HTTPStatusError(
            "409",
            request=MagicMock(),
            response=error_resp,
        )
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["coupons", "activate", "--all"])
        assert result.exit_code == 0
        assert "Skipped" in result.output

    @patch("ilidl.cli._get_client")
    def test_coupons_deactivate(self, mock_get_client):
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client

        runner = CliRunner()
        result = runner.invoke(cli, ["coupons", "deactivate", "c1"])
        assert result.exit_code == 0
        assert "Coupon deactivated" in result.output
        mock_client.deactivate_coupon.assert_called_once_with("c1")



"""Custom exceptions for ilidl."""


class ILidlError(Exception):
    """Base exception for all ilidl errors."""


class AuthError(ILidlError):
    """Authentication failed."""


class ReceiptParseError(ILidlError):
    """HTML receipt could not be parsed into structured data."""

    def __init__(self, message: str, raw_html: str) -> None:
        super().__init__(message)
        self.raw_html = raw_html



"""Tests for ilidl models."""

from datetime import datetime

from ilidl.models import Coupon, Discount, Item, Receipt, Store


class TestItem:
    def test_defaults(self):
        item = Item(name="Beans", price=2.74, vat_group="A")
        assert item.quantity == 1.0
        assert item.unit_price is None
        assert item.discounts == []

    def test_weight_item(self):
        item = Item(
            name="Loose Carrots",
            price=0.17,
            vat_group="A",
            quantity=0.24,
            unit_price=0.69,
        )
        assert item.quantity == 0.24
        assert item.unit_price == 0.69

    def test_item_with_discount(self):
        discount = Discount(description="Price Cut", amount=0.04)
        item = Item(
            name="Handwash",
            price=0.59,
            vat_group="B",
            discounts=[discount],
        )
        assert len(item.discounts) == 1
        assert item.discounts[0].amount == 0.04


class TestReceipt:
    def test_minimal(self):
        store = Store(
            id="GB1459",
            name="Llandaff",
            address="Station Road",
            postal_code="CF14 2FB",
            locality="Cardiff",
        )
        receipt = Receipt(
            id="123",
            date=datetime(2026, 3, 18, 19, 26, 50),
            store=store,
            items=[],
            total=12.38,
            currency="GBP",
        )
        assert receipt.payment_method is None
        assert receipt.vat_breakdown == {}


class TestCoupon:
    def test_defaults(self):
        coupon = Coupon(
            id="abc",
            title="Free item",
            start_date=datetime(2026, 3, 1),
            end_date=datetime(2026, 3, 31),
            is_activated=False,
        )
        assert coupon.description == ""
        assert coupon.image_url is None
